#define FALSE	0
#define TRUE	1

/*
 * rule typedefs
 */
typedef struct ex {
	int	type;
	union {
		int	resw;
		float	f;
	} u;
	struct ex	*left, *right;
} expr;

typedef struct cl {
	int		type;
	union {
		struct {
			float	min, max;
		} type1;
		struct {
			struct cl	*first, *alt;
		} type2;
	} u;
	struct cl	*nxt;
} condition;

typedef struct rl {
	condition	*conds;
	expr		*red, *green, *blue;
	struct rl	*nxt;
} rule;

/*
 * vector typedef
 */
typedef struct v {
	float	x, y, z;
} vector;

/*
 * macros for processing vectors
 */

/*
 * dprod
 *
 *	returns the dot product of a and b.
 */
#define	dprod(a, b)	((a).x * (b).x + (a).y * (b).y + (a).z * (b).z)

/*
 * xprod
 *
 *	returns the vector cross product of b and c in a
 */
#define xprod(a, b, c) { \
	(a).x = (b).y * (c).z - (b).z * (c).y; \
	(a).y = (b).z * (c).x - (b).x * (c).z; \
	(a).z = (b).x * (c).y - (b).y * (c).x; \
}

/*
 * normalise
 *
 *	normalise vector a
 */
#define	normalise(a) {\
	register float	_norm;\
\
	_norm = sqrt((double)dprod((a), (a)));\
	(a).x /= _norm;\
	(a).y /= _norm;\
	(a).z /= _norm;\
}

/*
 * vsub
 *
 *	assign the subtraction of c from b to a
 */
#define vsub(a, b, c) {\
	(a).x = (b).x - (c).x;\
	(a).y = (b).y - (c).y;\
	(a).z = (b).z - (c).z;\
}

/*
 * vadd
 *
 *	assign the addition of b and c into a
 */
#define vadd(a, b, c) {\
	(a).x = (b).x + (c).x;\
	(a).y = (b).y + (c).y;\
	(a).z = (b).z + (c).z;\
}

/*
 * smult
 *
 *	scalar multiply vector a by b
 */
#define	smult(a, b) {\
	(a).x *= (b);\
	(a).y *= (b);\
	(a).z *= (b);\
}

/*
 * miscellaneous macros
 */
#define	sqr(x)	((x) * (x))

#define clamp(a)        ((a) < 0.0 ? 0.0 : ((a) < 1.0 ? (a) : 1.0))

extern float	eval();
