#include <stdio.h>
#include <math.h>
#include "defs.h"
#include "y.tab.h"

static	float	xval, yval, zval;
static	float	dxval, dyval, dzval;

/*
 * eval
 *
 *	expression evaluation.
 */
float
eval(e)
	expr	*e;
{
	switch (e->type) {
	case FLOAT:
		return(e->u.f);
	case X:
		return(xval);
	case Y:
		return(yval);
	case Z:
		return(zval);
	case DX:
		return(dxval);
	case DY:
		return(dyval);
	case DZ:
		return(dzval);
	case PLUS:
		return(eval(e->left) + eval(e->right));
	case MINUS:
		return(eval(e->left) - eval(e->right));
	case DIV:
		return(eval(e->left) / eval(e->right));
	case MUL:
		return(eval(e->left) * eval(e->right));
	case ABS:
		return(fabs(eval(e->left)));
	default:
		fprintf(stderr, "eval: unknown expression type %d\n", e->type);
		exit(1);
	}
}

/*
 * evalconds
 *
 *	check the conditions to position (x, y, z) and the surface normal
 * (dx, dy, dz) at the current point. Returns non-zero if they match up.
 */
int
evalconds(conds, x, y, z, dx, dy, dz)
	condition	*conds;
	float		x, y, z;
	float		dx, dy, dz;
{
	int		failed;
	condition	*c;

	failed = FALSE;

	for (c = conds; !failed && c != (condition *)NULL; c = c->nxt)
		switch (c->type) {
		case X:
			if (c->u.type1.min > x || c->u.type1.max < x)
				failed = TRUE;
			break;
		case Y:
			if (c->u.type1.min > y || c->u.type1.max < y)
				failed = TRUE;
			break;
		case Z:
			if (c->u.type1.min > z || c->u.type1.max < z)
				failed = TRUE;
			break;
		case DX:
			if (c->u.type1.min > dx || c->u.type1.max < dx)
				failed = TRUE;
			break;
		case DY:
			if (c->u.type1.min > dy || c->u.type1.max < dy)
				failed = TRUE;
			break;
		case DZ:
			if (c->u.type1.min > dz || c->u.type1.max < dz)
				failed = TRUE;
			break;
		case OR:
			if (evalconds(c->u.type2.first, x, y, z, dx, dy, dz) == 0)
				failed = evalconds(c->u.type2.alt, x, y, z, dx, dy, dz);
			break;
		default:
			fprintf(stderr, "evalrule: unknown condition type %d\n", c->type);
			exit(1);
		}

	return(!failed);
}

/*
 * getrgb
 *
 *	calculate the rgb value for a given point according to the rules
 * in rulelist.
 */
getrgb(rules, x, y, z, dx, dy, dz, red, green, blue)
	rule	*rules;
	float	x, y, z;
	float	dx, dy, dz;
	float	*red, *green, *blue;
{
	int	failed;
	rule	*r;

	xval = x;
	yval = y;
	zval = z;

	dxval = dx;
	dyval = dy;
	dzval = dz;

	for (r = rules; r != (rule *)NULL; r = r->nxt) {
		if (evalconds(r->conds, x, y, z, dx, dy, dz)) {
			*red = eval(r->red);
			*green = eval(r->green);
			*blue = eval(r->blue);
			break;
		}
	}
}
