#include <stdio.h>
#include <math.h>

extern char	*strrchr();

/*
 * vector typedef
 */
typedef struct v {
	float	x, y, z;
} vector;

/*
 * macros for processing vectors
 */

/*
 * dprod
 *
 *	returns the dot product of a and b.
 */
#define	dprod(a, b)	((a).x * (b).x + (a).y * (b).y + (a).z * (b).z)

/*
 * xprod
 *
 *	returns the vector cross product of b and c in a
 */
#define xprod(a, b, c) { \
	(a).x = (b).y * (c).z - (b).z * (c).y; \
	(a).y = (b).z * (c).x - (b).x * (c).z; \
	(a).z = (b).x * (c).y - (b).y * (c).x; \
}

/*
 * normalise
 *
 *	normalise vector a
 */
#define	normalise(a) {\
	register float	_norm;\
\
	_norm = sqrt((double)dprod((a), (a)));\
	(a).x /= _norm;\
	(a).y /= _norm;\
	(a).z /= _norm;\
}

/*
 * vsub
 *
 *	assign the subtraction of c from b to a
 */
#define vsub(a, b, c) {\
	(a).x = (b).x - (c).x;\
	(a).y = (b).y - (c).y;\
	(a).z = (b).z - (c).z;\
}

/*
 * vadd
 *
 *	assign the addition of b and c into a
 */
#define vadd(a, b, c) {\
	(a).x = (b).x + (c).x;\
	(a).y = (b).y + (c).y;\
	(a).z = (b).z + (c).z;\
}

/*
 * smult
 *
 *	scalar multiply vector a by b
 */
#define	smult(a, b) {\
	(a).x *= (b);\
	(a).y *= (b);\
	(a).z *= (b);\
}

/*
 * miscellaneous macros
 */
#define	sqr(x)	((x) * (x))

#define clamp(a)        ((a) < 0.0 ? 0.0 : ((a) < 1.0 ? (a) : 1.0))

/*
 * calcnorm
 *
 *	calculate the normal at the triangle defined by the indexes
 * a, b, and c.
 */
calcnorm(n, a, b, c, z, xsize, ysize)
	vector	*n;
	int	a, b, c;
	float	*z;
	int	xsize, ysize;
{
	vector	v1, v2, v3, va, vb;
	float	fact;

	if (xsize > ysize)
		fact = xsize - 1;
	else
		fact = ysize - 1;

	v1.x = ((float)(a % xsize) / fact);
	v1.y = ((float)(a / ysize) / fact);
	v1.z = z[a];

	v2.x = ((float)(b % xsize) / fact);
	v2.y = ((float)(b / ysize) / fact);
	v2.z = z[b];

	v3.x = ((float)(c % xsize) / fact);
	v3.y = ((float)(c / ysize) / fact);
	v3.z = z[c];

	vsub(va, v2, v1);
	vsub(vb, v3, v2);

	xprod(*n, va, vb);

	normalise(*n);
}

/*
 * read in a heightfield file and generate a matching 
 * colour field file.
 */
main(ac, av)
	int	ac;
	char	*av[];
{
	FILE	*hf, *cf;
	int	xsize, ysize, x, y, i, j, ind;
	float	*zvals, xval, yval, zval, red, green, blue;
	char	buf[BUFSIZ], *p;
	vector	*norms, n;

	if (ac != 2) {
		fprintf(stderr, "gencolfield: usage gencol file.hf\n");
		exit(1);
	}

	if ((hf = fopen(av[1], "r")) == (FILE *)NULL) {
		fprintf(stderr, "gencolfield: can't open input file %s\n", av[1]);
		exit(1);
	}

	if (fscanf(hf, "HEIGHTFIELD %d %d ", &xsize, &ysize) == 0) {
		fread(&xsize, sizeof(xsize), 1, hf);
		ysize = xsize;
	}

	strcpy(buf, av[1]);
	p = strrchr(buf, '.');

	if (p == (char *)NULL || strcmp(p, ".hf") != 0) {
		fprintf(stderr, "gencolfield: heightfield file must end in .hf\n");
		exit(1);
	}

	strcpy(p, ".cf");

	if ((cf = fopen(buf, "w")) == (FILE *)NULL) {
		fprintf(stderr, "gencolfield: can't open output file %s\n", buf);
		exit(1);
	}

	fprintf(cf, "COLORFIELD %d %d\n", xsize, ysize);

	zvals = (float *)calloc(sizeof(float), xsize * ysize);

	if (zvals == (float *)NULL) {
		fprintf(stderr, "gencolfield: can't allocate memory for field\n");
		exit(1);
	}

	fread(zvals, sizeof(float), xsize * ysize, hf);

	norms = (vector *)calloc(sizeof(vector), xsize * ysize);

	if (norms == (vector *)NULL) {
		fprintf(stderr, "gencolfield: can't allocate memory for normals\n");
		exit(1);
	}

	for (i = 0; i < ysize - 1; i++)
		for (j = 0; j < xsize - 1 ; j++) {
			ind = i * xsize + j;

			calcnorm(&n, ind, ind + 1, ind + xsize + 1,
				zvals, xsize, ysize);

			vadd(norms[ind], norms[ind], n);
			vadd(norms[ind + 1], norms[ind + 1], n);
			vadd(norms[ind + xsize + 1], norms[ind + xsize + 1], n);                        
			calcnorm(&n, ind, ind + xsize + 1, ind + xsize,
				zvals, xsize, ysize);

			vadd(norms[ind], norms[ind], n);
			vadd(norms[ind + xsize + 1], norms[ind + xsize + 1], n);                        
			vadd(norms[ind + xsize], norms[ind + xsize], n);
		}

	for (i = 0; i < xsize * ysize; i++)
		normalise(norms[i]);

	for (y = 0; y != ysize; y++)
		for (x = 0; x != xsize; x++) {
			red = fabs(norms[y * xsize + x].x);
			green = fabs(norms[y * xsize + x].y);
			blue = fabs(norms[y * xsize + x].z);
			fwrite(&red, sizeof(float), 1, cf);
			fwrite(&green, sizeof(float), 1, cf);
			fwrite(&blue, sizeof(float), 1, cf);
		}

	fclose(hf);
	fclose(cf);
}
