#include <stdio.h>
#include <math.h>
#include "defs.h"

extern char	*strrchr();

rule		*rulelist;

int		lineno;

extern FILE	*yyin;

/*
 * calcnorm
 *
 *	calculate the normal at the triangle defined by the indexes
 * a, b, and c.
 */
calcnorm(n, a, b, c, z, xsize, ysize)
	vector	*n;
	int	a, b, c;
	float	*z;
	int	xsize, ysize;
{
	vector	v1, v2, v3, va, vb;
	float	fact;

	if (xsize > ysize)
		fact = xsize - 1;
	else
		fact = ysize - 1;

	v1.x = ((float)(a % xsize) / fact);
	v1.y = ((float)(a / ysize) / fact);
	v1.z = z[a];

	v2.x = ((float)(b % xsize) / fact);
	v2.y = ((float)(b / ysize) / fact);
	v2.z = z[b];

	v3.x = ((float)(c % xsize) / fact);
	v3.y = ((float)(c / ysize) / fact);
	v3.z = z[c];

	vsub(va, v2, v1);
	vsub(vb, v3, v2);

	xprod(*n, va, vb);

	normalise(*n);
}

/*
 * read in a heightfield file and generate a matching 
 * colour field file.
 */
main(ac, av)
	int	ac;
	char	*av[];
{
	FILE	*hf, *cf, *rf;
	int	xsize, ysize, x, y, i, j, ind, size;
	float	*zvals, xval, yval, zval, red, green, blue;
	float	minz, maxz;
	char	buf[BUFSIZ], *p;
	vector	*norms, n;

	if (ac != 2) {
		fprintf(stderr, "gencolf: usage gencol file.hf\n");
		exit(1);
	}

	if ((hf = fopen(av[1], "r")) == (FILE *)NULL) {
		fprintf(stderr, "gencolf: can't open input file %s\n", av[1]);
		exit(1);
	}

	if (fscanf(hf, "HEIGHTFIELD %d %d ", &xsize, &ysize) == 0) {
		fread(&xsize, sizeof(xsize), 1, hf);
		ysize = xsize;
	}

	strcpy(buf, av[1]);
	p = strrchr(buf, '.');

	if (p == (char *)NULL || strcmp(p, ".hf") != 0) {
		fprintf(stderr, "gencolf: heightfield file must end in .hf\n");
		exit(1);
	}

	strcpy(p, ".cf");

	if ((cf = fopen(buf, "w")) == (FILE *)NULL) {
		fprintf(stderr, "gencolf: can't open output file %s\n", buf);
		exit(1);
	}

	fprintf(cf, "COLORFIELD %d %d\n", xsize, ysize);

	zvals = (float *)calloc(sizeof(float), xsize * ysize);

	if (zvals == (float *)NULL) {
		fprintf(stderr, "gencolf: can't allocate memory for field\n");
		exit(1);
	}

	fread(zvals, sizeof(float), xsize * ysize, hf);

	minz = maxz = zvals[0];
	for (i = 1; i < ysize * xsize; i++)
		if (zvals[i] < minz)
			minz = zvals[i];
		else if (zvals[i] > maxz)
			maxz = zvals[i];
	
	printf("heightfield dimensions:\n");

	if (xsize > ysize) {
		size = xsize;
		printf("x: 0.0 1.0 y: 0.0 %f z %f %f\n", ysize / (float)xsize, minz, maxz);
	} else {
		size = ysize;
		printf("x: 0.0 %f y: 0.0 1.0 z %f %f\n", xsize / (float)ysize, minz, maxz);
	}

	norms = (vector *)calloc(sizeof(vector), xsize * ysize);

	if (norms == (vector *)NULL) {
		fprintf(stderr, "gencolf: can't allocate memory for normals\n");
		exit(1);
	}

	for (i = 0; i < ysize - 1; i++)
		for (j = 0; j < xsize - 1 ; j++) {
			ind = i * xsize + j;

			calcnorm(&n, ind, ind + 1, ind + xsize + 1,
				zvals, xsize, ysize);

			vadd(norms[ind], norms[ind], n);
			vadd(norms[ind + 1], norms[ind + 1], n);
			vadd(norms[ind + xsize + 1], norms[ind + xsize + 1], n);                        
			calcnorm(&n, ind, ind + xsize + 1, ind + xsize,
				zvals, xsize, ysize);

			vadd(norms[ind], norms[ind], n);
			vadd(norms[ind + xsize + 1], norms[ind + xsize + 1], n);                        
			vadd(norms[ind + xsize], norms[ind + xsize], n);
		}

	for (i = 0; i < xsize * ysize; i++)
		normalise(norms[i]);

	strcpy(p, ".rls");

	if ((yyin = fopen(buf, "r")) == (FILE *)NULL) {
		for (y = 0; y != ysize; y++)
			for (x = 0; x != xsize; x++) {
				red = fabs(norms[y * xsize + x].x);
				green = fabs(norms[y * xsize + x].y);
				blue = fabs(norms[y * xsize + x].z);
				fwrite(&red, sizeof(float), 1, cf);
				fwrite(&green, sizeof(float), 1, cf);
				fwrite(&blue, sizeof(float), 1, cf);
			}
	} else {
		yyparse();
		fclose(yyin);
		for (y = 0; y != ysize; y++)
			for (x = 0; x != xsize; x++) {
				getrgb(rulelist, x / (float)size, y / (float)size, zvals[y * xsize + x], norms[y * xsize + x].x, norms[y * xsize + x].y, norms[y * xsize + x].z, &red, &green, &blue);
				fwrite(&red, sizeof(float), 1, cf);
				fwrite(&green, sizeof(float), 1, cf);
				fwrite(&blue, sizeof(float), 1, cf);
			}

	}

	fclose(hf);
	fclose(cf);
}
